module Evoc
  class Scenario
    include Comparable, Logging

    attr_accessor :case_id,
                  :granularity,
                  :scenario_id,
                  :tx,
                  :algorithm,
                  :measures,
                  :aggregator,
                  :tx_index,
                  :tx_id,
                  :permutation,
                  :query,
                  :model_size,
                  :model_age,
                  :max_size,
                  :stats,
                  :opts

    def initialize(opts = Hash.new)
      logger.debug "Initialized new scenario with configuration: #{opts}"
      self.opts        = opts
      self.scenario_id = opts.hash

      self.tx_id = opts[:tx_id]
      self.model_age = opts[:model_age]
      opts.each do |attribute,value|
        self.send("#{attribute}=", value)
      end
    end


    ##
    # <=> defines how to compare two Query objects
    def <=> other
      return nil unless other.is_a?(Query)
      comparison = 0
      # first we compare the tx id
      if (self.tx_id <=> other.tx_id) == 0
        # if we also have the same query
        if (self.query.sort <=> other.query.sort) == 0
        # use history size as comparator
           comparison = (self.model_size <=> other.model_size)
        else
          # use the query
          comparison = (self.query.sort <=> other.query.sort)
        end
      else
        # use the tx id
        comparison = (self.tx_id <=> other.tx_id)
      end
      comparison
    end


    def to_h
      fields = %w(case_id granularity scenario_id tx_id tx_index tx_size query_size query_percentage expected_outcome_size model_size model_hours model_age max_size algorithm aggregator measures stats)
      hash = Hash.new
      fields.each do |key|
        value = self.method(key).call
        if value.is_a?(Array)
          hash[key] = value.join(',')
        elsif value.is_a?(Hash)
          hash.merge!(value)
        else
          hash[key] = value
        end
      end
      return hash
    end

    def recommendation?
        Evoc::RecommendationCache.recommendation_cached?(algorithm: self.algorithm,
                                                      query: self.query,
                                                      model_start: self.model_start,
                                                      model_end: self.model_end,
                                                      max_size: self.max_size)
    end

    def to_s
      self.opts
    end


    ##
    # CUSTOM SETTERS AND GETTERS
    #
    
    ##
    # @return [Evoc::Tx] the transaction of this scenario
    def tx
      tx = nil
      if !self.tx_id.nil?
        tx = Evoc::HistoryStore.base_history.get_tx(id: self.tx_id, id_type: :id)
      end
      return tx
    end

    ##
    # @return [Integer] the size of the query
    def query_size
      self.query.size
    end

    ##
    # @return [Float] the percentage of the whole history that is used in the model
    def model_percentage
      self.model_size.to_i == 0 ? 100 : ((self.model_size.to_f/self.tx_index)*100).round(2)
    end

    ##
    # Sets the model size in this scenario
    #
    # If set to 0 (zero), the maximum possible size is calculated
    # In this case, model age will be taken into account
    # 
    # @param [Integer] the size of the model
    def model_size=(size)
      if size.to_i == 0
        @model_size = (self.model_age.nil? ? self.tx_index : self.tx_index - self.model_age)
      else
        @model_size = size.to_i
      end
    end

    ##
    # @return [Integer] the index in the history where the model starts
    def model_start
      value = self.tx_index - self.model_size.to_i - self.model_age.to_i
      if value < 0
        raise ArgumentError, "Model out of bound"
      end
      return value
    end

    ##
    # @return [Integer] the index in the history where the model ends
    def model_end
      value = (self.tx_index - 1 - self.model_age.to_i)
      if value < 0
        raise ArgumentError, "The model end index was negative (model_size:#{self.model_size}, tx_index:#{self.tx_index})"
      elsif value < self.model_start
        raise ArgumentError, "The model end was before the model start (start: #{self.model_start}, end: #{value})"
      end
      return value
    end

    ##
    # @return [Integer] the time between the first and last transaction in the model
    def model_hours
        model_end_tx = Evoc::HistoryStore.base_history.get_tx(id: self.model_end)
        model_start_tx = Evoc::HistoryStore.base_history.get_tx(id: self.model_start)
        return TimeDifference.between(model_start_tx.date,model_end_tx.date).in_hours
    end

    ##
    # query_percentage
    def query_percentage
      total_items = self.query.size.to_i + self.expected_outcome.size
      (self.query.size.to_f/total_items*100).round
    end

    
    ##
    # max_size
    def max_size=(value)
      if !value.nil?
        if value.respond_to?(:to_i)
          @max_size = value.to_i
        else
          raise ArgumentError.new, "value for max_size could not be converted to integer, value was: #{value}"
        end
      end
    end

    ##
  	# custom setter for query
    def query=(query)
      if !query.nil?
        raise ArgumentError, "Query must be an array" unless query.is_a?(Array)
        @query = query
        # convert to list of integers
        @query.map!(&:to_i) 
        if !self.tx.nil?
          if @query.size >= (self.tx.size)
            $stderr.warn "The query was larger than or equal to the size of the transaction (Qs: #{@query.size}, Tx size: #{self.tx.size})"
          end
        end
        if @query.empty?
          raise ArgumentError, "The query was empty"
        end
      end
    end

    ##
    # Custom setter for measures
    def measures=(measures)
      if !measures.nil?
        # internally, all interestingness measures are 
        # represented as :m_measurename (symbols), so we concatinate
        # 'm_' and symbolize the list of given measures
        if measures.is_a?(String)
          measures = measures.split(',')
        end
        @measures = measures.map {|m| (/\Am_/ =~ m).nil? ? ('m_'+m).to_sym : m.to_sym}
      end
    end

    ##
    # @return [Array] the list of expected items
  	def expected_outcome
      expected_outcome = (self.tx.items - self.query)
      if expected_outcome.empty?
        logger.warn "The expected outcome was empty"
      end
      return expected_outcome
  	end
    
    ##
    # @return [Integer] the size of the expected outcome
  	def expected_outcome_size
      self.expected_outcome.size
  	end

    ##
    # @return [String] the history index of this scenarios transaction
    def tx_index
      self.tx.index
    end

    ##
    # @return [Integer] the size of this scenarios transaction
    def tx_size
      self.tx.size
    end

    def stats
      time_start = Time.now
      history = Evoc::HistoryStore.get_history(self.model_start,self.model_end,self.max_size)
      relevant_transactions = Set.new
      relevant_items = Set.new
      relevant_ages = []
      avg_age_of_relevant_transactions = 0
      avg_size_of_relevant_transactions = 0
      files_changed = Set.new
      num_methods_changed = 0
      num_new_items = 0
      # @avg_size_of_relevant_transactions = 0
      # mean_age_of_relevant
      # media_age_of_relevant
      # ratio_new_items
      self.query.each do |item|
        if change = history.int_2_name[item]
          change = change.split(':')
          files_changed << change[0]
          if change.size > 1
            num_methods_changed = num_methods_changed + 1
          end
          indexes_of_previous_changes = history.transactions_of(item, identifier: :index)
          if new_item = (indexes_of_previous_changes.size == 0)
            num_new_items = num_new_items + 1
          else
            indexes_of_previous_changes.each do |tx_index|
              relevant_so_far = relevant_transactions.size
              relevant_transactions << tx_index
              new_relevant = (relevant_transactions.size > relevant_so_far)
              if new_relevant
                age = (self.tx_index - tx_index)
                relevant_ages << age
                tx = history.get_tx(id: tx_index,id_type: :index)
                avg_size_of_relevant_transactions = avg_size_of_relevant_transactions + tx.size
                avg_age_of_relevant_transactions = avg_age_of_relevant_transactions + age
              end
            end
          end
        else
          num_new_items = num_new_items + 1
        end
      end

      num_relevant_transactions = relevant_transactions.size

      time_end = Time.now
      time_generate_stats = TimeDifference.between(time_start,time_end).in_seconds.round(8)
      {time_generate_stats: time_generate_stats,
       num_files_changed: files_changed.size,
       num_methods_changed: num_methods_changed,
       num_new_items: num_new_items,
       num_relevant_transactions: num_relevant_transactions,
       median_age_of_relevant_transactions: relevant_ages.median,
       avg_age_of_relevant_transactions: num_relevant_transactions == 0 ? nil : (avg_age_of_relevant_transactions/num_relevant_transactions).to_f,
       avg_size_of_relevant_transactions: num_relevant_transactions == 0 ? nil : (avg_size_of_relevant_transactions/num_relevant_transactions).to_f}
    end


  end
end
