require_relative 'cli_helper'

module EvocCLI
  class Analyze < Thor
    class_option :transactions, :aliases => '-t', :type => :string, :required => false, :desc => "Path to change-history"
    class_option :logger_level, type: :string, default: 'error', desc: "The severity level to output to debug.log"
    class_option :granularity, type: :string, default: 'mixed', desc: "The granularity level of imported items from history"

    desc "evolution","Outputs where the items of all transactions previously changed."
    def evolution
      a = Evoc::Analyze.new(options)
      a.evolution
    end

    desc "commits","Outputs the items touched so far and the moving average of items per transactions for every transaction in the history"
    def commits
      a = Evoc::Analyze.new(options)
      a.commits
    end

    method_option :number, aliases: '-n', type: :numeric, default: 1000, desc: "The number of rules to calculate measures for"
    desc "measure_values","Empirically investigate the range of interestingness measures"
    def measure_values
      a = Evoc::Analyze.new(options)
      a.measure_values
    end

    desc "measure_ranges","Empirically investigate the range of interestingness measures"
    def measure_ranges
      a = Evoc::Analyze.new(options)
      a.measure_ranges
    end

    desc "aggregator_range","Generate example output of aggregators given a list of values to be aggregated"
    def aggregator_range
      a = Evoc::Analyze.new(options)
      a.aggregator_range
    end

    desc "uniqueness [options]","Run uniqueness analysis on TRANSACTIONS"
    long_desc <<-LONGDESC
    Run an analysises of how many unique transactions there are in TRANSACTIONS.

    The analysis is based on the order present in TRANSACTIONS. A transaction
    is considered unique if no previous transactions are a superset of the
    given transaction.
    LONGDESC
    def uniqueness
      a = Evoc::Analyze.new(options)
      $stdout.puts a.uniqueness
    end

    desc "all", "Outputs a collection of stats about the input history"
    def all
      Evoc::Analyze.new(options).all
    end

    desc "avg_changes_per_file", "Output the average number of commits that each file is present in"
    def avg_changes_per_file
      a = Evoc::Analyze.new(options)
      $stdout.puts __method__
      $stdout.puts a.average_changes_per_file
    end

    desc "num_commits", "Output the number of commits in the input history"
    def num_commits
      a = Evoc::Analyze.new(options)
      $stdout.puts __method__
      $stdout.puts a.num_commits
    end

    desc "time_span [options]","Dump the time span between first and last commits in history"
    def time_span
      a = Evoc::Analyze.new(options)
      unit = 'years'
      $stdout.puts "#{__method__}_#{unit}"
      $stdout.puts a.time_span(unit: unit)
    end

    desc "avg_time_between_commits [options]","Dump the average time between each commit"
    def avg_time_between_commits
      a = Evoc::Analyze.new(options)
      unit = 'hours'
      $stdout.puts "#{__method__}_#{unit}"
      $stdout.puts a.average_time_between_commits(unit: unit)
    end

    desc "num_unique_files [options]","Dump the number of unique files mentioned in the history"
    def num_unique_files
      a = Evoc::Analyze.new(options)
      $stdout.puts __method__
      $stdout.puts a.num_unique_files
    end

    method_option :top, :aliases => '-n', :type => :numeric, default: 10, :desc => "How many files to return"
    desc "file_frequency [options]","Return the X most frequent files"
    def file_frequency
      a = Evoc::Analyze.new(options)
      a.file_frequency
    end


    method_option :group, :aliases => '-g', :type => :boolean, default: false, :desc => "If the commits should be grouped by size"
    desc "commit_size", "Dumps the commits sizes to stdout"
    def commit_size
      a = Evoc::Analyze.new(options)
      a.commit_size
    end

    desc "avg_commit_size", "Dumps the average commits size of history"
    def avg_commit_size
      a = Evoc::Analyze.new(options)
      $stdout.puts __method__
      $stdout.puts a.average_commit_size
    end

    desc "create_dict", ""
    def create_dict
      a = Evoc::Analyze.new(options)
      a.create_dict
    end

    desc "recommendation_meta",""
    def recommendation_meta

      int i = 1
      ARGF.each_line do |line|

        rec = JSON.parse(line)

        tx_id = rec["tx_id"]
        algorithm = rec["algorithm"]
        ap = rec["average_precision1000"]
        rank = nil
        if !ap.nil?
          if ap > 0
            rat = ap.rationalize
            # the rank is only valid if we obtained a 1 in the numerator
            if rat.numerator == 1
              rank = rat.denominator
            end
          end
        else
          $stderr.puts "AP was nil for #{tx_id}"
        end
        applicable = !rec["rules"].empty?
        positive_recommendation = !rank.nil?
        # find the confidence of the correct item
        expected_outcome = nil
        confidence_correct = nil
        rank_one = nil
        confidence_incorrect = nil
        discernibility = nil
        ordered_rec = unique_strongest_consequents(rec["rules"])
        if ordered_rec.size > 0
          if positive_recommendation
            expected_outcome = ordered_rec[rank-1][0]
            confidence_correct = ordered_rec[rank-1][1]
          else
            rank_one = ordered_rec[0][0]
            confidence_incorrect = ordered_rec[0][1]
          end
          discernibility = (ordered_rec.group_by{|(_,confidence)| confidence}.size).to_f/ordered_rec.size
        end


        # build output hash
        hash = {tx_id: tx_id,
          algorithm: algorithm,
          ap: ap,
          rank: rank,
          applicable: applicable,
          positive_recommendation: positive_recommendation,
          expected_outcome: expected_outcome,
          confidence_correct: confidence_correct,
          rank_one: rank_one,
          confidence_incorrect: confidence_incorrect,
          discernibility: discernibility}
        begin
          puts hash.to_json
        rescue JSON::GeneratorError => e
          $stderr.puts "Failed to convert hash to JSON, error was #{e}\n the hash was \n#{hash}\n\n"
          next
        end
      $stderr.print "#{i} lines processed"
      i += 1
      end
    end

    private
    def unique_strongest_consequents(rules)
      selected_consequents = Hash.new{|h,k| h[k] = 0}
      rules.each do |rule|
        lhs = rule["lhs"]
        rhs = rule["rhs"]
        confidence = rule["measures"]["m_confidence"]
        if !confidence.nil?
          if confidence.to_r.to_f > selected_consequents[rhs]
            selected_consequents[rhs] = confidence.to_r.to_f
          end
        else
          $stderr.puts "Confidence was nil for rule #{lhs} -> #{rhs}"
        end
      end
      return selected_consequents.sort_by {|k,v| -v}
    end
  end
end

